from datetime import datetime
from enum import Enum
import json
import logging
from pydantic import BaseModel, Field, validate_arguments
import requests
import time
from typing import Any, Dict, List, Optional
from urllib.parse import quote_plus, urlparse


class FilePath(BaseModel):
    name: str
    path: str
    uid: Optional[str]


class FieldType(str, Enum):
    text = "STRING"
    number = "NUMBER"
    bool = "BOOL"
    timestamp = "TIMESTAMP"
    entity = "ENTITY"


class FieldSpec(BaseModel):
    field: str  # This is the field's name
    type: FieldType
    expression: Optional[str]
    path: Optional[List[str]]
    isList: Optional[bool]


class InferredFieldType(Enum):
    text = {"expression": "`text_string-dev-1`(File.text)", "type": "STRING"}
    number = {"expression": "`text_number-dev-1`(File.text)", "type": "NUMBER"}
    timestamp = {"expression": "`text_date-dev-1`(File.text)", "type": "TIMESTAMP"}
    checkbox = {"expression": "checkbox(File.text)", "type": "STRING"}
    signature = {"expression": "region_signature(File.text)", "type": "STRING"}
    table = {
        "expression": "entity_one_many(File.text)",
        "type": "ENTITY",
        "isList": True,
    }

    @property
    def expr(self):
        return self.value["expression"]

    @classmethod
    def match_trainer(cls, trainer: str):
        ret = None
        for x in cls:
            if trainer in x.expr:
                assert ret is None, "Matched multiple trainers: %s and %s" % (ret, x)
                ret = x
        assert ret is not None, "Unknown trainer: %s" % (trainer)
        return ret

    @validate_arguments
    def build_field_spec(self, field_name: str, path: List[str] = []) -> FieldSpec:
        return FieldSpec(field=field_name, path=path, **self.value)


class InvalidRequest(Exception):
    pass


class APIError(Exception):
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return "%s: %s" % (self.response.status_code, self.response.content)


class IQLError(Exception):
    pass


class ResourceType(str, Enum):
    fc = "fc"
    dc = "dc"
    ec = "ec"
    collection = "collection"


@validate_arguments
def parse_date(s: str) -> datetime:
    return datetime.strptime(s, "%Y-%m-%d")


class Impira:
    def __init__(
        self, org_name, api_token, base_url="https://app.impira.com", ping=True
    ):
        self.org_url = urljoin(base_url, "o", org_name)
        self.api_url = urljoin(self.org_url, "api/v2")
        self.headers = {"X-Access-Token": api_token}

        if ping:
            try:
                self._ping()
            except Exception as e:
                raise InvalidRequest(
                    "Failed to ping Impira API. Please check your org name and API token. Full error: %s"
                    % str(e)
                )

    @validate_arguments
    def upload_files(self, collection_id: Optional[str], files: List[FilePath]):
        local_files = len([f for f in files if urlparse(f.path).scheme in ("", "file")])
        if local_files > 0 and local_files != len(files):
            raise InvalidRequest(
                "All files must be local or URLs, but not a mix (%d/%d were local)"
                % (local_files, len(files))
            )
        elif local_files > 0:
            return self._upload_multipart(collection_id, files)
        else:
            return self._upload_url(collection_id, files)

    @validate_arguments
    def get_collection_uid(self, collection_name: str):
        resp = self.query('@__system::collections[uid] Name="%s"' % (collection_name))[
            "data"
        ]
        if len(resp) == 0:
            return None

        uids = [x["uid"] for x in resp]
        if len(uids) > 1:
            raise InvalidRequest(
                "Found multiple collections with name '%s': %s"
                % (collection_name, ", ".join(uids))
            )

        return uids[0]

    @validate_arguments
    def update(self, collection_id: str, data: List[Dict[str, Any]]):
        resp = requests.patch(
            self._build_resource_url("fc", collection_id),
            headers=self.headers,
            json={"data": data},
        )

        if not resp.ok:
            raise APIError(resp)

        return resp.json()["uids"]

    @validate_arguments
    def add_files_to_collection(self, collection_id: str, file_ids: List[str]):
        resp = requests.post(
            self._build_resource_url("ec", "file_collection_contents"),
            headers=self.headers,
            json={
                "data": [
                    {"file_uid": u, "collection_uid": collection_id} for u in file_ids
                ]
            },
        )

        if not resp.ok:
            raise APIError(resp)

    @validate_arguments
    def get_app_url(self, resource_type: ResourceType, resource_id: str) -> str:
        return self._build_resource_url(
            resource_type, resource_id, api=False, use_async=False
        )

    @validate_arguments
    def create_collection(self, collection_name: str):
        existing = self.get_collection_uid(collection_name)
        if existing is not None:
            raise InvalidRequest(
                "Collection with name '%s' already exists at %s"
                % (collection_name, urljoin(self.org_url, "fc", existing))
            )

        # Create collection is implemented as an empty insert
        resp = requests.post(
            self._build_resource_url("collection", collection_name),
            headers=self.headers,
            json={"data": []},
        )

        if not resp.ok:
            raise APIError(resp)

        uid_list = resp.json()["uids"]
        assert (
            not uid_list
        ), "Expected empty uid list while creating a collection, but received: %s" % (
            ", ".join(uid_list)
        )

        return self.get_collection_uid(collection_name)

    @validate_arguments
    def create_field(self, collection_id: str, field_spec: FieldSpec):
        resp = requests.post(
            urljoin(
                self.api_url, "schema/ecs/file_collections::%s/fields" % (collection_id)
            ),
            headers=self.headers,
            json=dict(field_spec),
        )

        if not resp.ok:
            raise APIError(resp)

    @validate_arguments
    def create_fields(self, collection_id: str, field_specs: List[FieldSpec]):
        resp = requests.post(
            urljoin(
                self.api_url, "schema/ecs/file_collections::%s/fields" % (collection_id)
            ),
            headers=self.headers,
            json=[dict(field_spec) for field_spec in field_specs],
        )

        if not resp.ok:
            raise APIError(resp)

    @validate_arguments
    def create_inferred_field(
        self,
        collection_id: str,
        field_name: str,
        inferred_field_type: InferredFieldType,
        path: List[str] = [],
    ):
        field_spec = inferred_field_type.build_field_spec(
            field_name=field_name, path=path
        )
        return self.create_field(collection_id, field_spec)

    @validate_arguments
    def poll_for_results(self, collection_id: str, uids: List[str] = None):
        uid_filter = (
            "and in(uid, %s)" % (", ".join(['"%s"' % u for u in uids])) if uids else ""
        )
        query = """
        @`file_collections::%s`
            File.IsPreprocessed=true and __system.IsProcessed=true
            %s
            [.: __resolve(.)]""" % (
            collection_id,
            uid_filter,
        )

        cursor = None

        must_see = set(uids)
        while True:
            resp = self.query(query, mode="poll", cursor=cursor, timeout=60)
            for d in resp["data"] or []:
                if d["action"] != "insert":
                    continue
                yield d["data"]

                uid = d["data"]["uid"]
                if len(must_see) > 0:
                    assert uid in must_see, "Broken uid filter (%s not in %s)" % (
                        uid,
                        must_see,
                    )
                    must_see.remove(uid)

            cursor = resp["cursor"]

            if len(must_see) == 0:
                break

    @validate_arguments
    def query(
        self, query: str, mode: str = "iql", cursor: str = None, timeout: int = None
    ):
        args = {"query": query}
        if cursor is not None:
            args["cursor"] = cursor

        if timeout is not None:
            args["timeout"] = timeout

        resp = requests.post(
            urljoin(self.api_url, mode),
            headers=self.headers,
            json=args,
        )
        if not resp.ok:
            raise APIError(resp)

        d = resp.json()
        if "error" in d and d["error"] is not None:
            raise IQLError(d["error"])

        return resp.json()

    def _ping(self):
        self.query("@files limit:0")

    @validate_arguments
    def _upload_multipart(self, collection_id: Optional[str], files: List[FilePath]):
        for i in range(60):
            for f in files:
                if f.uid is not None:
                    raise InvalidRequest(
                        "Unsupported: specifying a UID in a multi-part file upload (%s)"
                        % (f.uid)
                    )

            files_body = [
                t
                for f in files
                for t in [
                    ("file", open(f.path, "rb")),
                    ("data", json.dumps(_build_file_object(f.name, None, f.uid))),
                ]
            ]
            resp = requests.post(
                self._build_collection_url(collection_id, use_async=True),
                headers=self.headers,
                files=tuple(files_body),
            )
            if resp.status_code == 429:
                logging.warning(
                    "Sleeping for 2 seconds and then retrying multi-part upload..."
                )
                time.sleep(2)
            elif not resp.ok:
                raise APIError(resp)
            else:
                return resp.json()["uids"]

    @validate_arguments
    def _upload_url(self, collection_id: Optional[str], files: List[FilePath]):
        resp = requests.post(
            self._build_collection_url(collection_id, use_async=True),
            headers=self.headers,
            json={"data": [_build_file_object(f.name, f.path, f.uid) for f in files]},
        )
        if not resp.ok:
            raise APIError(resp)

        return resp.json()["uids"]

    @validate_arguments
    def _upload_url_collection_name(self, collection_name: str, files: List[FilePath]):
        resp = requests.post(
            self._build_collection_url(collection_id, use_async=True),
            headers=self.headers,
            json={"data": [_build_file_object(f.name, f.path, f.uid) for f in files]},
        )
        if not resp.ok:
            raise APIError(resp)

        return resp.json()["uids"]

    @validate_arguments
    def _build_collection_url(self, collection_id: Optional[str], use_async=False):
        if collection_id is not None:
            return self._build_resource_url("fc", collection_id, use_async)
        else:
            return self._build_resource_url("files", use_async)

    @validate_arguments
    def _build_resource_url(
        self, resource_type: ResourceType, resource_id: str, api=True, use_async=False
    ):
        base_url = urljoin(
            self.api_url if api else self.org_url,
            resource_type,
            quote_plus(resource_id),
        )
        if use_async:
            base_url = base_url + "?async=1"
        return base_url


def _build_file_object(name, path=None, uid=None):
    ret = {"File": {"name": name}}
    if path is not None:
        ret["File"]["path"] = path
    if uid is not None:
        ret["uid"] = uid
    return ret


def urljoin(*parts):
    return "/".join([x.lstrip("/") for x in parts])



from concurrent.futures import ThreadPoolExecutor
import pathlib
import requests
from shutil import copyfile
from uuid import uuid4

from ..tools.impira import Impira
from ..config import get_logger
from ..schema import record_to_schema
from ..types import DocManifest
from .utils import add_datadir_arg

log = get_logger("snapshot")


def build_parser(subparsers, parent_parser):
    parser = subparsers.add_parser(
        "snapshot",
        help="Snapshot an Impira collection (fields, documents, and labels)",
        parents=[parent_parser],
    )
    Impira.add_arguments(parser)

    add_datadir_arg(parser)

    parser.add_argument(
        "--collection",
        default=None,
        type=str,
        required=True,
        help="uid of the collection to snapshot",
    )

    parser.add_argument(
        "--download-files",
        "-s",
        default=False,
        action="store_true",
        help="Download the files to disk.",
    )

    parser.add_argument(
        "--original-names",
        default=False,
        action="store_true",
        help="Use original filenames (without concatenating a uid). This will fail if two files in the collection have the same name",
    )

    parser.set_defaults(func=main)
    return parser


def download_file_to(url, path):
    r = requests.get(url)
    with open(path, "wb") as f:
        f.write(r.content)


def main(args):
    impira = Impira(config=Impira.Config(**vars(args)))
    workdir = pathlib.Path(args.data).joinpath(
        "capture", args.collection + "-" + str(uuid4())[:4]
    )

    schema, records = impira.snapshot(
        collection_uid=args.collection, use_original_filenames=args.original_names
    )

    log.info("Downloading %d files to %s", len(records), workdir)
    workdir.mkdir(parents=True, exist_ok=True)

    if args.download_files:
        with ThreadPoolExecutor(max_workers=args.parallelism) as t:
            [
                _
                for _ in t.map(
                    lambda r: download_file_to(r["url"], workdir.joinpath(r["name"])),
                    records,
                )
            ]

        docs = [{"fname": r["name"], "record": r["record"]} for r in records]
    else:
        docs = [
            {"fname": r["name"], "url": r["url"], "record": r["record"]}
            for r in records
        ]

    manifest = DocManifest(
        doc_schema=schema,
        docs=docs,
    )
    with open(workdir.joinpath("manifest.json"), "w") as f:
        f.write(manifest.json(indent=2))

    log.info("Documents and labels have been written to directory '%s'", workdir)



from ..config import get_logger


class Tool(object):
    def _log(self):
        if not (hasattr(self, "_cached_log") and self._cached_log is not None):
            self._cached_log = get_logger(type(self).__name__.lower())
        return self._cached_log



import os
import tempfile


def environ_or_required(key, default=None):
    return (
        {"default": os.environ.get(key, default)}
        if os.environ.get(key, default) is not None
        else {"required": True}
    )


def add_datadir_arg(parser):
    parser.add_argument(
        "--data",
        "-d",
        help="Directory to save documents.",
        **environ_or_required(
            "IMPIRA_DATA_DIR", os.path.join(tempfile.gettempdir(), "impira-cli")
        ),
    )



import boto3
from collections import OrderedDict
import logging
import os
import pathlib
from pydantic import BaseModel, validate_arguments
from uuid import uuid4
from typing import Any, Dict, List, Optional, Tuple, Type
import time
from trp import Document, SelectionElement, Word

from ..cmd.utils import environ_or_required
from ..schema import record_to_schema, schema_to_model
from ..types import (
    BBox,
    CheckboxLabel,
    DocData,
    NumberLabel,
    TextLabel,
    TimestampLabel,
)
from .tool import Tool


def convert_textract_bbox(bbox, page_num) -> BBox:
    return BBox(
        top=bbox.top,
        left=bbox.left,
        height=bbox.height,
        width=bbox.width,
        page=page_num,
    )


def parse_field(field, page_num, log=None):
    if field.value is None:
        # We don't know the bounding box for the value, so there is no point
        # trying to define the field
        return None

    if len(set([type(c) for c in field.value.content])) != 1:
        if log:
            log.warning(
                "Field '%s' on page %d with value '%s' has multiple types: %s. Skipping...",
                field.key.text,
                page_num + 1,
                field.value.text,
                field.value.content,
            )
        return None

    if isinstance(field.value.content[0], SelectionElement):
        bbox = field.value.geometry.boundingBox
        assert field.value.text in ("SELECTED", "NOT_SELECTED")
        value = CheckboxLabel(
            value=field.value.text == "SELECTED",
            location=convert_textract_bbox(field.value.geometry.boundingBox, page_num),
        )
    elif isinstance(field.value.content[0], Word):
        # Textract only reports values as "text" types. There is no number parsing (as far as I can tell).
        # With typed systems like Impira, we rely on their additional information while creating fields
        # to create a more specific type.
        bbox = field.value.geometry.boundingBox
        value = TextLabel(
            value=field.value.text,
            location=convert_textract_bbox(field.value.geometry.boundingBox, page_num),
        )
    else:
        assert False, "Unknown value type: %s" % (field.value.content[0])

    return value


def doc_to_record(log, doc):
    fields = OrderedDict()
    for page_num, page in enumerate(doc.pages):
        for field in page.form.fields:
            value = parse_field(field, page_num)
            if value is not None:
                fields[field.key.text] = value

    # Construct a pydantic object for this record
    T = schema_to_model(record_to_schema(fields))
    return T(**fields)


class Textract(Tool):
    class Config(BaseModel):
        s3_bucket: Optional[str]
        s3_prefix: str

    @staticmethod
    def add_arguments(parser):
        parser.add_argument(
            "--s3-bucket",
            default=None,
            type=str,
            help="an existing s3 bucket to use for staging files. if not specified, impira cli will try to find an existing one or create one",
        )
        parser.add_argument(
            "--s3-prefix", **environ_or_required("TEXTRACT_S3_BUCKET", "")
        )

    @validate_arguments
    def __init__(self, config: Config):
        self.config = config

    def _init_bucket(self):
        if self.config.s3_bucket is not None:
            return

        log = self._log()
        session = boto3.Session()
        s3 = boto3.client("s3")

        matching_bucket = None
        for bucket in s3.list_buckets()["Buckets"]:
            if "impira-cli-staging-area" not in bucket["Name"]:
                continue
            location = s3.get_bucket_location(Bucket=bucket["Name"])[
                "LocationConstraint"
            ]
            if location is None or location == session.region_name:
                matching_bucket = bucket["Name"]
                break

        if matching_bucket is None:
            log.info(
                "No 'impira-cli-staging-area' bucket exists, going to try creating one..."
            )
            matching_bucket = "impira-cli-staging-area-" + str(uuid4())
            s3.create_bucket(Bucket=matching_bucket)
            log.info("Created bucket s3://%s", matching_bucket)
        else:
            log.info("Using existing bucket s3://%s", matching_bucket)

        self.config.s3_bucket = matching_bucket

    @validate_arguments
    def textract_document(self, fname: pathlib.Path, forms=True, tables=False):
        self._init_bucket()

        log = self._log()

        s3 = boto3.client("s3")
        key = os.path.join(self.config.s3_prefix, str(uuid4()), fname.name)
        s3.upload_file(str(fname), self.config.s3_bucket, key)

        feature_types = []
        if forms:
            feature_types.append("FORMS")
        if tables:
            feature_types.append("TABLES")

        try:
            textract = boto3.client("textract")
            job_info = textract.start_document_analysis(
                DocumentLocation={
                    "S3Object": {"Bucket": self.config.s3_bucket, "Name": key}
                },
                FeatureTypes=feature_types,
            )

            job_id = job_info["JobId"]

            response = None
            while response is None or response["JobStatus"] == "IN_PROGRESS":
                if response is not None:
                    log.info("Sleeping for 5 seconds...")
                    time.sleep(5)

                response = textract.get_document_analysis(JobId=job_id)

            assert response["JobStatus"] == "SUCCEEDED"

            log.debug("Reading textract results")
            pages = [response]
            while True:
                next_token = response.get("NextToken", None)
                if next_token is None:
                    break
                response = textract.get_document_analysis(
                    JobId=job_id, NextToken=next_token
                )
                pages.append(response)
        finally:
            log.debug("Cleaning up file on S3")
            s3.delete_object(Bucket=self.config.s3_bucket, Key=key)

        return Document(pages)

    @validate_arguments
    def process_document(self, fname: pathlib.Path, forms=True, tables=False):
        doc = self.textract_document(fname, forms, tables)

        return doc_to_record(self._log(), doc)
